from datetime import date, timedelta
from typing import Optional

from dateutil.relativedelta import relativedelta


class SpecialDate(object):
    @classmethod
    def _offset_by_one_week(cls, date_require: date, offset: int) -> date:
        """
        获取某天往前看的最近的星期n
        :param date_require 从哪一天开始往前看
        :param offset 1 周日 2 周六 3 周五 4 周四 5 周三 6 周二 7 周一
        """
        today = date.today() if not date_require else date_require
        offset = today.weekday() + offset
        return today - timedelta(days=offset)

    @classmethod
    def last_sunday(cls, date_require: date) -> date:
        """
        获取某天上一周的周日
        :param date_require: 某天
        :return: date
        """
        return cls._offset_by_one_week(date_require, 1)

    @classmethod
    def sunday_before_n_weeks(cls, require_day: date, weeks: int) -> date:
        """
        获取n周前的周日
        :param require_day: 某天
        :param weeks: 周数
        :return: date
        """
        return cls.last_sunday(require_day) - relativedelta(weeks=weeks)


if __name__ == "__main__":
    date_util = SpecialDate.sunday_before_n_weeks(date.today(), 0)
    print(date_util)



import os
import threading
import time
import uuid
from typing import Optional

import redis
from redis.client import Pipeline

from lightning_fast.tools.base_tools.ip_tools import IpTools

REDIS_LOCK_PREFIX = "REDIS_READ_WRITE_LOCK"


class RedisReadWriteLock:
    def __init__(self, conn, lock_name, acquire_timeout=10, lock_timeout=1000):
        self.conn = conn
        self.lock_name = lock_name
        self.acquire_timeout = acquire_timeout
        self.lock_timeout = lock_timeout
        self.identifier = None

    @property
    def acquire_success(self):
        return self.identifier is not None

    def __enter__(self):
        self.identifier = self.acquire_lock(
            self.conn, self.lock_name, self.acquire_timeout, self.lock_timeout
        )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.identifier is not None:
            self.release_lock(self.conn, self.lock_name, self.identifier)

    @classmethod
    def combine_lock_key(cls, lock_name):
        return f"{REDIS_LOCK_PREFIX}:" + lock_name

    @classmethod
    def acquire_lock(
        cls,
        conn: redis.StrictRedis,
        lock_name: str,
        acquire_timeout: int,
        lock_timeout: int,
    ) -> Optional[str]:
        """
        :param conn: redis连接
        :param lock_name: 锁名字
        :param acquire_timeout: 拿锁超时时间(秒)
        :param lock_timeout: 锁自然过期时间(秒)
        :return: 锁的唯一值
        """
        # 拿到本机网络ip
        ip = IpTools.get_ip()
        # 拿到本程序进程id
        pid = os.getpid()
        # 拿到线程id
        thread_name = threading.current_thread().name
        # 生成一个唯一用来删除时识别的值
        identifier = f"{ip}-{pid}-{thread_name}-{uuid.uuid4()}"
        # 生成一个key
        lock_key = cls.combine_lock_key(lock_name)
        # 根据获取超时时间设定结束获取时间
        end = time.time() + acquire_timeout
        while time.time() < end:
            # 使用 setnx 本操作如果已经有key会返回0, 否则返回1
            if conn.setnx(lock_key, identifier):
                # 设置过期时间
                conn.expire(lock_key, lock_timeout)
                return identifier
            # 如果发现没有设置过期时间(-1)或者没有key(-2)或者非严格模式(None)
            else:
                ttl_time = conn.ttl(lock_key)
                if ttl_time is None or (-1 <= ttl_time <= 0):
                    conn.expire(lock_key, lock_timeout)
            time.sleep(0.001)

        return None

    @classmethod
    def release_lock(
        cls, conn: redis.StrictRedis, lock_name: str, identifier: str
    ) -> bool:
        """
        :param conn: redis连接
        :param lock_name: 锁名字
        :param identifier: 拿锁的时候创建的唯一值, 用来识别是否可以删掉这个锁, 别删错了
        :return: 返回是否释放成功
        """
        # 创建redis pipeline
        pipe: Pipeline = conn.pipeline(True)
        # 生成一个key
        lock_key = cls.combine_lock_key(lock_name)
        while True:
            try:
                # break transaction once lock has been changed, use watch watch result in pipeline.
                pipe.watch(lock_key)
                # 如果获取到的key是需要识别的值
                lock_value = pipe.get(lock_key)
                if isinstance(lock_value, bytes):
                    lock_value = lock_value.decode()
                if lock_value == identifier:
                    # 增加命令
                    pipe.multi()
                    # 删除当前的key
                    pipe.delete(lock_key)
                    # 执行
                    pipe.execute()
                    # 返回释放成功
                    return True

                # execute when identifier not equal. 当没有拿到的时候把watch释放
                pipe.unwatch()
                # 直接结束
                break
            except redis.exceptions.WatchError as e:
                # 如果发现中间有人把lock给改了, 直接结束
                print(e)
                return False
        return False



from collections import deque, OrderedDict
from copy import deepcopy
import joblib
import numpy as np
import pandas as pd
from tqdm import tqdm


class CommonLabelEncoder(object):
    def __init__(self, transform_shift=0):
        """
        此处选择编码移动数量，比如想从1开始则shift=1
        """
        self.hash_dict = OrderedDict()
        self.inverse_dict = OrderedDict()
        self.counter = 0
        self.transform_shift = transform_shift

    @property
    def labels_in_order(self):
        return list(self.hash_dict.keys())

    def fit(self, words, data_type="one", reset=True):
        """
        fit_type代表输入是否是一个嵌套结构
        """
        if reset:
            self.hash_dict = OrderedDict()
            self.inverse_dict = OrderedDict()
        self.counter = 0
        if isinstance(words, np.ndarray) or isinstance(words, pd.Series):
            if words.size == 0:
                return
        else:
            if not words:
                return
        if data_type == "one":
            self._one_depth_fit(words)
        elif data_type == "nested":
            self._nested_fit(words)

    def _one_depth_fit(self, words):
        counter = 0
        for word in words:
            if word not in self.hash_dict:
                self.hash_dict[word] = counter
                self.inverse_dict[counter] = word
                counter += 1
        self.counter = counter

    @classmethod
    def judge_custom_support_iterable(cls, item):
        return (
            isinstance(item, list)
            or isinstance(item, pd.Series)
            or isinstance(item, np.ndarray)
        )

    def _nested_fit(self, words):
        if not self.judge_custom_support_iterable(words):
            raise ValueError(f"输入类型不能为{type(words)}")
        bar = tqdm()
        counter = 0
        cache_queue = deque()
        cache_queue.append(words)
        while cache_queue:
            current_item = cache_queue.popleft()
            if self.judge_custom_support_iterable(current_item):
                for item in current_item:
                    cache_queue.append(item)
            else:
                if current_item not in self.hash_dict:
                    self.hash_dict[current_item] = counter
                    self.inverse_dict[counter] = current_item
                    bar.update(1)
                    counter += 1
        self.counter = counter

    def transform(self, new_words, data_type="one"):
        if data_type == "one":
            return self._common_transform(
                new_words, self.hash_dict, self.transform_shift
            )
        elif data_type == "nested":
            return self._common_nested_transform(
                new_words, self.hash_dict, self.transform_shift
            )
        else:
            raise ValueError(f"data_type不能为{data_type}")

    def inverse_transform(self, encoded_words, data_type="one"):
        if data_type == "one":
            return self._common_transform(
                encoded_words, self.inverse_dict, self.transform_shift, inverse=True
            )
        elif data_type == "nested":
            return self._common_nested_transform(
                encoded_words, self.inverse_dict, self.transform_shift, inverse=True
            )
        else:
            raise ValueError(f"data_type不能为{data_type}")

    @classmethod
    def _common_transform(cls, words, mapping, shift, inverse=False):
        result = []
        for word in words:
            if not inverse:
                code = mapping.get(word, -1)
                code = code if code == -1 else code + shift
            else:
                code = mapping.get(word - shift, -1)
                code = str(code)
            result.append(code)
        return result

    def _common_nested_transform(self, words, mapping, shift, inverse=False):
        bar = tqdm()
        result = deepcopy(words)
        cache = deque()
        cache.append(result)
        while cache:
            current_item = cache.popleft()
            if self.judge_custom_support_iterable(current_item):
                if len(current_item) > 0:
                    current_item_children = (
                        current_item[0]
                        if not isinstance(current_item, pd.Series)
                        else current_item.iloc[0]
                    )
                    if not self.judge_custom_support_iterable(current_item_children):
                        current_item[:] = self._common_transform(
                            current_item, mapping, shift, inverse=inverse
                        )
                        bar.update(1)
                    else:
                        for sub_item in current_item:
                            cache.append(sub_item)
        return result

    def print_encoder(self):
        print(self.hash_dict)

    def save_encoder(self, save_path):
        joblib.dump(self, save_path)

    @classmethod
    def load_encoder(cls, load_path):
        return joblib.load(load_path)



from pymongo import MongoClient

from lightning_fast.mongodb.clients.client_descriptor.base.client_descriptor_base import ClientDescriptorBase


class SyncMongodbClient(ClientDescriptorBase):
    def _create_client(self) -> MongoClient:
        return MongoClient(self.mongodb_url, connect=False)



from datetime import timedelta, date
from typing import Optional

import pandas as pd
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta

from lightning_fast.tools.time_tools.special_date import SpecialDate


class SpecialDateRange:
    @classmethod
    def get_week_bins(cls, start_date, end_date, weeks=1):
        # 将给定的起止时间调整为开始时间向前最近的周一，结束时间向后最近的周日
        # 返回范围内的长度为weeks周的周1-周日列表
        start_week_range = pd.date_range(
            parse(start_date) - timedelta(days=6), parse(start_date)
        )
        nearest_monday = start_week_range[start_week_range.weekday == 0][0]
        end_week_range = pd.date_range(
            parse(end_date), parse(end_date) + timedelta(days=6)
        )
        nearest_sunday = end_week_range[end_week_range.weekday == 6][0]
        return pd.DataFrame(
            {
                "start_date": pd.date_range(
                    nearest_monday, nearest_sunday, freq=f"{weeks * 7}D", closed="left"
                )
                .strftime("%Y-%m-%d")
                .to_list(),
                "end_date": (
                    pd.date_range(
                        nearest_monday,
                        nearest_sunday,
                        freq=f"{weeks * 7}D",
                        closed="left",
                    )
                    + timedelta(days=weeks * 7 - 1)
                )
                .strftime("%Y-%m-%d")
                .to_list(),
            }
        )

    @classmethod
    def get_week_range_by_date(cls, some_date):
        """
        当前日期所属周范围
        :param some_date: 指定日期
        :return: 当前周1到周日
        """
        first_date = some_date - timedelta(days=some_date.weekday())
        last_date = some_date + timedelta(days=7 - some_date.weekday() - 1)
        return first_date.strftime("%Y-%m-%d"), last_date.strftime("%Y-%m-%d")

    @classmethod
    def generate_week_index(cls, start_date, end_date):
        """
        返回从开始日期到结束日期的按周分时间段
        :param start_date: 开始时间
        :param end_date: 结束时间
        :return: [(周1, 周日), ...]
        """
        date_range = pd.date_range(start_date, end_date)
        week_index = pd.Series(
            date_range.map(lambda x: x.isocalendar()[0] * 10000 + x.isocalendar()[1])
        ).rank(method="dense")
        date_week = pd.DataFrame([date_range, week_index.rank(method="dense")]).T
        bin_starts = (
            date_week.groupby(1)
            .first()[0]
            .map(lambda x: x.strftime("%Y-%m-%d"))
            .values.tolist()
        )
        bin_ends = (
            date_week.groupby(1)
            .last()[0]
            .map(lambda x: x.strftime("%Y-%m-%d"))
            .values.tolist()
        )
        return list(zip(bin_starts, bin_ends))

    @classmethod
    def get_nature_week_ranges_before_month_and_week(
        cls,
        date_require: Optional[date] = None,
        week_count_range: int = 1,
        months: int = 1,
        weeks: int = 0,
    ):
        """
        months与weeks定义需要获取的范围宽度，比如months=1, weeks=1则获取的区间长度都是1月加一周
        week_count_range定义需要从date_require前数多少周获取这个范围。
        上述参数返回值为
        [
            [last sunday - months - weeks , last sunday],
            [last sunday - months - weeks - 7, last sunday - 7],
            [last sunday - months - weeks - 14, last sunday - 14],
        ]
        :param date_require: 以哪天为基准
        :param week_count_range: 前数几周
        :param months: n月前
        :param weeks: n周前
        :return: 二维数组
        """
        if not date_require:
            date_require = date.today()
        position_date = SpecialDate.last_sunday(date_require)
        return [
            [
                (
                    position_date
                    - timedelta(days=count * 7)
                    - relativedelta(months=months)
                    - relativedelta(weeks=weeks)
                ),
                (position_date - timedelta(days=count * 7)),
            ]
            for count in range(week_count_range)
        ]


if __name__ == "__main__":
    print(SpecialDateRange.get_week_bins("2019-11-01", "2020-11-01", 1))
    print(SpecialDateRange.generate_week_index("2019-11-01", "2020-11-01"))



from lightning_fast.tools.time_tools.special_date import SpecialDate
from lightning_fast.tools.time_tools.special_date_points import SpecialDatePoints
from lightning_fast.tools.time_tools.special_date_range import SpecialDateRange


__all__ = ["SpecialDate", "SpecialDatePoints", "SpecialDateRange"]



import os
from functools import partial

import pandas as pd
from matplotlib import pyplot as plt
import seaborn as sns
from sklearn.preprocessing import LabelEncoder
from tqdm import tqdm

from tests.test_utils import LocalCToolsCompiler
from tests.time_statistic.old_encoder_statistic import (
    LabelEncoderComparison,
    SingleSpendStatistic,
)
from tests.test_config import ROOT_PATH

if len(list(os.listdir(LocalCToolsCompiler.target_dir))) == 0:
    LocalCToolsCompiler().compile_local_c_tools()

encoders = LocalCToolsCompiler.import_encoders()
CLabelEncoder = encoders.LabelEncoder
OneDStringVector = encoders.OneDStringVector

if __name__ == "__main__":
    tmp_directory = ROOT_PATH / "tmp"
    tmp_directory.mkdir(exist_ok=True, parents=True)
    test_chars = ["赵", "钱", "孙", "李", "周", "吴", "郑", "王"]
    total_test_words = LabelEncoderComparison.generate_words(test_chars, 0, 7, 600000)
    result = []
    for i in tqdm(
        range(
            100000,
            600000,
            int((600000 - 100000) / 50),
        )
    ):
        sklearn_test_words = total_test_words[:i]
        sklearn_label_encoder = LabelEncoder()
        c_test_words = OneDStringVector(sklearn_test_words)
        c_label_encoder = CLabelEncoder()
        sk_fit_time = SingleSpendStatistic.get_function_ms(
            partial(sklearn_label_encoder.fit, sklearn_test_words)
        )
        sk_transform_time = SingleSpendStatistic.get_function_ms(
            partial(sklearn_label_encoder.transform, sklearn_test_words)
        )
        c_fit_time = SingleSpendStatistic.get_function_ms(
            partial(c_label_encoder.encode_1d, c_test_words, 8)
        )
        c_transform_time = SingleSpendStatistic.get_function_ms(
            partial(c_label_encoder.transform_1d, c_test_words, 8)
        )
        print(
            f"sk_fit_time: {sk_fit_time}, c_fit_time: {c_fit_time}, sk_transform_time: {sk_transform_time}, c_transform_time: {c_transform_time}"
        )
        result.append([
            i, sk_fit_time, sk_transform_time, c_fit_time, c_transform_time
        ])

    columns = [
        "word_count",
        "sk_fit_time",
        "sk_transform_time",
        "c_fit_time",
        "c_transform_time",
    ]
    result_df = pd.DataFrame(result, columns=columns)
    plt.close()
    result_df.plot(x="word_count")
    plt.savefig(tmp_directory / f"new_compare.png")



import re


def underline_to_camel(name: str) -> str:
    """
    这是一个奖下划线转为驼峰的方法
    :param name: 待转换字符串
    :return: 转换后字符串
    """
    under_pat = re.compile(r"_([0-9a-z])")
    return under_pat.sub(lambda x: x.group(1).upper(), name)



import os
from ctypes import CDLL, POINTER, c_int, c_char_p, c_char, cast
from functools import lru_cache

import numpy as np

from lightning_fast.c_structs.custom_structure import UniqueWords
from lightning_fast.encoders.c_encoders.c_utils import DataConverter
from lightning_fast.config import ROOT_PATH


class CEncoderPath(object):
    def __init__(self):
        self.source_dir = ROOT_PATH / "c_dynamic_library"

    @property
    @lru_cache()
    def all_libs(self):
        return os.listdir(self.source_dir)

    @property
    @lru_cache()
    def label_encoder_path(self):
        for file in self.all_libs:
            if file.startswith("liblabel_encoder"):
                return self.source_dir / file
        raise FileNotFoundError(f"Can not find lib file 'liblabel_encoder.*'")


class CLabelEncoder(object):
    """
    必须要自己手动调用free_encoder来释放产生的encoder内存
    """

    def __init__(self):
        self._c_label_encoder = None
        self._c_encoder_module = None
        self.init_c_encoder_module()

    def init_c_encoder_module(self):
        self._c_encoder_module = CDLL(CEncoderPath().label_encoder_path)
        self._c_encoder_module.get_encoder_by_string.argtypes = [c_char_p, c_char]
        self._c_encoder_module.get_encoder_by_string.restype = POINTER(UniqueWords)
        self._c_encoder_module.free_unique_words.argtypes = [
            POINTER(POINTER(UniqueWords))
        ]
        self._c_encoder_module.print_unique_words_summary.argtypes = [
            POINTER(POINTER(UniqueWords))
        ]
        self._c_encoder_module.save_encoder.argtypes = [
            POINTER(POINTER(UniqueWords)),
            c_char_p,
        ]
        self._c_encoder_module.load_encoder.argtypes = [c_char_p]
        self._c_encoder_module.load_encoder.restype = POINTER(UniqueWords)
        self._c_encoder_module.encode_words_by_string.argtypes = [
            POINTER(POINTER(UniqueWords)),
            c_char_p,
            c_int,
            c_char,
        ]
        self._c_encoder_module.encode_words_by_string.restype = POINTER(c_int)

    def fit(self, words):
        c_words = DataConverter.string_list_to_c_string(words)
        self._c_label_encoder = self._c_encoder_module.get_encoder_by_string(
            c_words, c_char("\t".encode("utf-8"))
        )

    def transform(self, words):
        c_words = DataConverter.string_list_to_c_string(words)
        result = self._c_encoder_module.encode_words_by_string(
            self._c_label_encoder, c_words, len(words), c_char("\t".encode("utf-8"))
        )
        buffer_as_ctypes_array = cast(result, POINTER(c_int * len(words)))[0]
        return np.frombuffer(buffer_as_ctypes_array, np.int32)

    def save_encoder(self, save_path):
        self._c_encoder_module.save_encoder(
            self._c_label_encoder, c_char_p(str(save_path).encode("utf-8"))
        )

    def load_encoder(self, load_path):
        self._c_label_encoder = self._c_encoder_module.load_encoder(
            c_char_p(str(load_path).encode("utf-8"))
        )

    def free_encoder(self):
        self._c_encoder_module.free_unique_words(self._c_label_encoder)

    @property
    def c_module_empty(self):
        return not self._c_encoder_module

    @property
    def c_encoder_empty(self):
        return not self._c_label_encoder

    def print_encoder(self):
        self._c_encoder_module.print_unique_words_summary(self._c_label_encoder)



from lightning_fast.tools.dask_tools.dask_reader import DaskReader
from lightning_fast.tools.dask_tools.dask_config_controller import DaskConfigController
from lightning_fast.tools.dask_tools.dask_computer import DaskComputer

__all__ = [
    "DaskReader",
    "DaskComputer",
    "DaskConfigController"
]



import json

from lightning_fast.tools.json_tools.decoders.mongodb_doc_decoder import (
    MongodbDocDecoder,
)


class JsonDecoder:
    @classmethod
    def decode_mongodb(cls, mongodb_string):
        return json.loads(
            mongodb_string, cls=MongodbDocDecoder,
        )



import os
import pathlib
import logging
from typing import Optional


class DirectoryDescriptor(object):
    """
    此描述器当路径存在的时候不处理，当路径不存在的时候，会创建路径
    """
    def __init__(self, directory: Optional[pathlib.Path]):
        self.directory = directory
        if directory is not None:
            self._create_if_not_existed(directory)

    def __set__(self, instance, value: pathlib.Path):
        if value is not None:
            self._create_if_not_existed(value)
            self.directory = value

    def __get__(self, instance, owner):
        return self.directory

    @classmethod
    def _create_if_not_existed(cls, directory):
        if os.path.exists(directory):
            logging.info(f"{directory}--已存在")
        else:
            logging.info(f"创建--{directory}")
            directory.mkdir(exist_ok=True, parents=True)



import subprocess
import sys

from lightning_fast.config import ROOT_PATH

source_dir = ROOT_PATH / ".." / "c_source"
target_dir = ROOT_PATH / "c_dynamic_library"
build_dir = source_dir / "cmake-build-c-dynamic-library"
build_dir.mkdir(parents=True, exist_ok=True)

subprocess.check_call(
    [
        "cmake",
        "..",
        f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={target_dir}",
        f"-DPYTHON_EXECUTABLE={sys.executable}",
        f"-DCMAKE_BUILD_TYPE=Release"
    ],
    cwd=build_dir,
)
subprocess.check_call("make", cwd=build_dir)


if __name__ == "__main__":
    from lightning_fast.c_dynamic_library.encoders import LabelEncoder
    from lightning_fast.c_dynamic_library.encoders import OneDStringVector

    from tests.time_statistic.old_encoder_statistic import LabelEncoderComparison

    total_test_words = LabelEncoderComparison.generate_words(
        ["赵", "钱", "孙", "李", "周", "吴", "郑", "王"], 0, 7, 100
    )
    input_string_list = OneDStringVector(total_test_words)
    encoder = LabelEncoder()
    encoder.encode_1d(input_string_list, 1)
    print(encoder.label_map)



from lightning_fast.tools.nest_object_tools.name_converter import NameConverter
from lightning_fast.tools.nest_object_tools.convert_functions import underline_to_camel


__all__ = ["NameConverter", "underline_to_camel"]



class YamlExample:
    @staticmethod
    def print():
        print(
            f"""
development:
  mongodbs:
    mongodb_1:
      "host": ""
      "user_name": ""
      "password": ""
      "auth": ""
    mongodb_2:
      "uri": ""

  external_api:
    api_1:
      "url": ""

    api_2:
      "url": ""


prod:
  mongodbs:
    mongodb_1:
      "host": ""
      "user_name": ""
      "password": ""
      "auth": ""
    mongodb_2:
      "uri": ""

  external_api:
    api_1:
      "url": ""

    api_2:
      "url": ""

common:
  SENTRY_DSN:

  SMTP_SERVER:
        """
        )


if __name__ == "__main__":
    YamlExample.print()



import unittest
from lightning_fast.tools.meta_tools import SingletonMeta


class TestCLabelEncoder(unittest.TestCase):
    def test_singleton(self):
        class TmpClass(metaclass=SingletonMeta):
            def __init__(self, counter):
                counter.append(1)

        # 第一次传参因为此时没有对象创建，所以应该会进入__init__, 此时list长度应该为1
        counter_list_1 = []
        TmpClass(counter_list_1)
        self.assertTrue(len(counter_list_1) == 1)
        # 第二次传参因为此时已经有对象被创建, 所以不会进入__init__, 此时list长度应该为2
        counter_list_2 = []
        TmpClass(counter_list_2)
        self.assertTrue(len(counter_list_2) == 0)


if __name__ == "__main__":
    unittest.main()



import os
import unittest
from lightning_fast.config import ROOT_PATH
from lightning_fast.encoders.common_encoders.common_label_encoder import CommonLabelEncoder


class TestCommonLabelEncoder(unittest.TestCase):

    test_strings = ["中文1", "中文2", "中文2", "符号，", "符号。", "符号；", "", "english"]
    test_encode_result = [0, 1, 1, 2, 3, 4, 5, 6]
    tmp_encoder_path = ROOT_PATH / "tmp" / "test_common_label_encoder.tpl"

    def test_encoder_fit(self):
        # 测试是否能正常生成encoder
        encoder = CommonLabelEncoder()
        encoder.fit(self.test_strings)
        encoder.print_encoder()

    def test_encoder_transform(self):
        # 测试是否能正常transform
        encoder = CommonLabelEncoder()
        encoder.fit(self.test_strings)
        transform_result = encoder.transform(self.test_strings)
        self.assertListEqual(self.test_encode_result, transform_result)

    def test_encoder_save(self):
        # 测试能够正常保存
        self.assertFalse(os.path.exists(self.tmp_encoder_path))
        encoder = CommonLabelEncoder()
        encoder.save_encoder(self.tmp_encoder_path)
        self.assertTrue(os.path.exists(self.tmp_encoder_path))
        os.remove(self.tmp_encoder_path)
        self.assertFalse(os.path.exists(self.tmp_encoder_path))

    def test_encoder_load(self):
        # 测试能否正常加载
        encoder = CommonLabelEncoder()
        encoder.save_encoder(self.tmp_encoder_path)
        self.assertTrue(self.tmp_encoder_path)
        encoder = CommonLabelEncoder()
        encoder.load_encoder(self.tmp_encoder_path)
        encoder.fit(self.test_strings)
        transform_result = encoder.transform(self.test_strings)
        self.assertListEqual(self.test_encode_result, transform_result)
        os.remove(self.tmp_encoder_path)


if __name__ == '__main__':
    unittest.main()



from abc import ABCMeta, abstractmethod

import redis

from lightning_fast.config_factory.config_constructor import Config


class PoolDescriptorBase(metaclass=ABCMeta):
    """
    此描述符其宿主必须有config属性，此属性是Config的实例，
    则此描述符的行为是：
    初始化时传入config
    赋值时生成连接字符串
    引用时创建连接
    """

    def __init__(self, redis_name: str):
        self.redis_url = None
        self.redis_client = None
        self.redis_config = None
        self.redis_name = redis_name

    def __set_name__(self, owner, name):
        """
        创建此描述符实例的时候, 直接拿宿主的config属性复制到描述符里边。
        """
        try:
            config: Config = getattr(owner, "config")
        except AttributeError:
            raise AttributeError(f"{owner} must have 'config' attribute.")
        if not isinstance(config, Config):
            raise TypeError(f"config must be instance of Config")
        try:
            self.redis_config = getattr(config, "redis")
        except AttributeError:
            raise AttributeError(f"config must have 'redis' attribute.")
        self._set_redis(self.redis_name)

    def __set__(self, instance, mongodb_name: str):
        """
        赋值时直接将配置中的字段名转换为mongodb连接字符串
        """
        self._set_redis(mongodb_name)

    def __get__(self, instance, owner):
        """
        引用的时候才会创建client
        """
        if self.redis_client is None:
            self.redis_client = self._create_pool()
        return self.redis_client

    @abstractmethod
    def _create_pool(
        self,
    ) -> redis.ConnectionPool:
        pass

    def _set_redis(self, redis_name: str):
        if redis_name in self.redis_config:
            conf = self.redis_config[redis_name]
            redis_host = conf.get("host", "127.0.0.1")
            redis_port = conf.get("port", "6379")
            redis_db = conf.get("db", 0)
            self.redis_url = f"redis://{redis_host}:{redis_port}/{redis_db}"
        else:
            self.redis_url = None



from abc import ABCMeta, abstractmethod
from typing import Union

import motor.motor_asyncio
from pymongo import MongoClient

from lightning_fast.config_factory.config_constructor import Config


class ClientDescriptorBase(metaclass=ABCMeta):
    """
    此描述符其宿主必须有config属性，此属性是Config的实例，
    则此描述符的行为是：
    初始化时传入config
    赋值时生成连接字符串
    引用时创建连接
    """
    def __init__(self, mongodb_name: str):
        self.mongodb_url = None
        self.mongodb_client = None
        self.mongodbs_config = None
        self.mongodb_name = mongodb_name

    def __set_name__(self, owner, name):
        """
        创建此描述符实例的时候, 直接拿宿主的config属性复制到描述符里边。
        """
        try:
            config = getattr(owner, "config")
        except AttributeError:
            raise AttributeError(f"{owner} must have 'config' attribute.")
        if not isinstance(config, Config):
            raise TypeError(f"config must be instance of Config")
        try:
            self.mongodbs_config = getattr(config, "mongodbs")
        except AttributeError:
            raise AttributeError(f"config must have 'mongodbs' attribute.")
        self._set_mongodb(self.mongodb_name)

    def __set__(self, instance, mongodb_name: str):
        """
        赋值时直接将配置中的字段名转换为mongodb连接字符串
        """
        self._set_mongodb(mongodb_name)

    def __get__(self, instance, owner):
        """
        引用的时候才会创建client
        """
        if self.mongodb_url is None:
            return self.mongodb_url
        if self.mongodb_client is None:
            self.mongodb_client = self._create_client()
        return self.mongodb_client

    @abstractmethod
    def _create_client(
        self,
    ) -> Union[MongoClient, motor.motor_asyncio.AsyncIOMotorClient]:
        pass

    def _set_mongodb(self, mongodb_name: str):
        if mongodb_name in self.mongodbs_config:
            conf = self.mongodbs_config[mongodb_name]
            if "uri" in conf:
                self.mongodb_url = conf["uri"]
            elif "user_name" in conf and "password" in conf:
                self.mongodb_url = (
                    f"mongodb://{conf['user_name']}:{conf['password']}@{conf['host']}/"
                )
            else:
                self.mongodb_url = None
        else:
            self.mongodb_url = None



from ctypes import (
    Structure,
    c_int, c_char_p,
)

from lightning_fast.c_structs.uthash import UT_hash_handle


class UniqueWords(Structure):
    pass


UniqueWords._fields_ = [
    ("word", c_char_p),
    ("value", c_int),
    ("hh", UT_hash_handle),
]



import os
import json
from datetime import datetime, timedelta
from functools import lru_cache
from random import randint, seed
import joblib
from sklearn.preprocessing import LabelEncoder
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from tqdm import tqdm

from lightning_fast.config import ROOT_PATH
from lightning_fast.encoders import CommonLabelEncoder, CLabelEncoder


class BaseStatisticResult(object):
    def __init__(
        self,
        name,
        word_count,
        create_spend,
        fit_spend,
        transform_spend,
        save_spend,
        load_spend,
    ):
        self.name = name
        self.word_count = word_count
        self.create_spend = create_spend
        self.fit_spend = fit_spend
        self.transform_spend = transform_spend
        self.save_spend = save_spend
        self.load_spend = load_spend

    def to_dict(self):
        result = vars(self)
        for key in result:
            if isinstance(result[key], timedelta):
                result[key] = result[key].total_seconds()
        return result

    def __str__(self):
        result = vars(self)
        for key in result:
            result[key] = str(result[key])
        return str(json.dumps(result, indent=2))

    def __repr__(self):
        return self.__str__()


class SingleSpendStatistic(object):
    def __init__(self, test_words, tmp_file_path, encoder):
        self.test_words = test_words
        self.tmp_file_path = tmp_file_path
        self.encoder = encoder
        self.encoder_instance = None

    def get_statistic(self):
        return BaseStatisticResult(
            self.encoder_type,
            len(self.test_words),
            self.get_function_ms(self._create),
            self.get_function_ms(self._fit),
            self.get_function_ms(self._transform),
            self.get_function_ms(self._save),
            self.get_function_ms(self._load),
        )

    @classmethod
    def get_function_ms(cls, func):
        start = datetime.now()
        func()
        end = datetime.now()
        return end - start

    @property
    @lru_cache()
    def encoder_type(self):
        if self.encoder.__name__ == "LabelEncoder":
            return "scikit"
        elif self.encoder.__name__ == "CommonLabelEncoder":
            return "common"
        elif self.encoder.__name__ == "CLabelEncoder":
            return "c"
        else:
            raise TypeError("Wrong encoder type!")

    def _create(self):
        if self.encoder_type == "scikit":
            self.encoder_instance = LabelEncoder()
        elif self.encoder_type == "common":
            self.encoder_instance = CommonLabelEncoder()
        else:
            self.encoder_instance = CLabelEncoder()

    def _fit(self):
        self.encoder_instance.fit(self.test_words)

    def _transform(self):
        self.encoder_instance.transform(self.test_words)

    def _save(self):
        if self.encoder_type == "scikit":
            joblib.dump(self.encoder_instance, self.tmp_file_path)
        else:
            self.encoder_instance.save_encoder(self.tmp_file_path)

    def _load(self):
        if self.encoder_type == "scikit":
            joblib.load(self.tmp_file_path)
        elif self.encoder_type == "c":
            self.encoder_instance.load_encoder(self.tmp_file_path)
            self.encoder_instance.free_encoder()
        else:
            self.encoder_instance.load_encoder(self.tmp_file_path)

    def _del_tmp_file(self):
        os.remove(self.tmp_file_path)


class LabelEncoderComparison(object):
    tmp_c_label_encoder_path = ROOT_PATH / "tmp/c_label_encoder.tpl"
    tmp_common_label_encoder_path = ROOT_PATH / "tmp/common_label_encoder.tpl"
    tmp_scikit_label_encoder_path = ROOT_PATH / "tmp/scikit_label_encoder.tpl"

    def __init__(
        self, chars, word_min_length, word_max_length, word_count, test_words=None
    ):
        if not test_words:
            self.test_words = test_words
        else:
            self.test_words = self.generate_words(
                chars, word_min_length, word_max_length, word_count
            )

    def process(self):
        scikit_label_encoder_result = SingleSpendStatistic(
            self.test_words, self.tmp_scikit_label_encoder_path, LabelEncoder
        ).get_statistic()
        common_label_encoder_result = SingleSpendStatistic(
            self.test_words, self.tmp_common_label_encoder_path, CommonLabelEncoder
        ).get_statistic()
        c_label_encoder_result = SingleSpendStatistic(
            self.test_words, self.tmp_c_label_encoder_path, CLabelEncoder
        ).get_statistic()
        return (
            scikit_label_encoder_result,
            common_label_encoder_result,
            c_label_encoder_result,
        )

    @classmethod
    def generate_words(cls, chars, word_min_length, word_max_length, word_count):
        seed(222)
        test_words = []
        for i in range(word_count):
            sub_word = ""
            for j in range(randint(word_min_length, word_max_length)):
                sub_word += chars[randint(0, len(chars) - 1)]
            test_words.append(sub_word)
        return test_words


class VarCompareResult(object):
    bins = 50

    def __init__(
        self, chars, word_min_length, word_max_length, word_min_count, word_max_count
    ):
        self.chars = chars
        self.word_min_length = word_min_length
        self.word_max_length = word_max_length
        self.word_min_count = word_min_count
        self.word_max_count = word_max_count
        self.total_test_words = LabelEncoderComparison.generate_words(
            chars, word_min_length, word_max_length, word_max_count
        )

    def get_var_result(self):
        result = []
        for i in tqdm(
            range(
                self.word_min_count,
                self.word_max_count,
                int((self.word_max_count - self.word_min_count) / self.bins),
            )
        ):
            sub_results = LabelEncoderComparison(
                self.chars,
                self.word_min_length,
                self.word_max_length,
                i,
                self.total_test_words[0:i],
            ).process()
            result += [sub_result.to_dict() for sub_result in sub_results]
        return result

    def draw_pic(self, tmp_directory):
        columns = [
            "create_spend",
            "fit_spend",
            "transform_spend",
            "save_spend",
            "load_spend",
        ]
        result_df = pd.DataFrame(self.get_var_result())
        result_df.loc[:, "word_count"] = result_df.loc[:, "word_count"].astype(int)
        for column in columns:
            result_df.loc[:, column] = result_df.loc[:, column].astype(float)
        for column in columns:
            plt.close()
            sns.lineplot(result_df["word_count"], result_df[column], result_df["name"])
            plt.savefig(tmp_directory / f"{column}.png")


if __name__ == "__main__":
    test_chars = ["赵", "钱", "孙", "李", "周", "吴", "郑", "王"]
    test_word_min_length = 0
    test_word_max_length = 7
    test_word_count = 600000
    # single_compare_result = LabelEncoderComparison(
    #     test_chars, test_word_min_length, test_word_max_length, test_word_count
    # ).process()
    # print(single_compare_result)
    var_compare_result = VarCompareResult(
        test_chars, test_word_min_length, test_word_max_length, 100000, test_word_count
    )
    test_tmp_directory = ROOT_PATH / "tmp/label_encoder_comparison"
    test_tmp_directory.mkdir(exist_ok=True, parents=True)
    var_compare_result.draw_pic(test_tmp_directory)



import os
import re
import sys
import traceback

from setuptools import setup, find_packages, Extension
from setuptools.command.install import install
from setuptools.command.egg_info import egg_info
from setuptools.command.build_ext import build_ext, _build_ext
import subprocess

# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}
# command = "sh prepare_env_source.sh"
# process = subprocess.Popen(command, shell=True, cwd=".")


class CustomInstallCommand(install):
    """Custom install setup to help run shell commands (outside shell) before installation"""

    def run(self):
        custom_command = "sh prepare_env_source.sh"
        # custom_process = subprocess.Popen(custom_command, shell=True, cwd=".")
        # custom_process.wait()
        try:
            subprocess.check_call(custom_command, shell=True, cwd=".")
        except subprocess.CalledProcessError:
            print("Not have cmake")
        install.run(self)


class CustomEggInfoCommand(egg_info):
    """Custom install setup to help run shell commands (outside shell) before installation"""

    def run(self):
        egg_info.run(self)
        custom_command = "sh prepare_env_source.sh"
        # custom_process = subprocess.Popen(custom_command, shell=True, cwd=".")
        # custom_process.wait()
        try:
            subprocess.check_call(custom_command, shell=True, cwd=".")
        except subprocess.CalledProcessError:
            print("Not have cmake")


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        """Build extensions in build directory, then copy if --inplace"""
        self.inplace = 0
        _build_ext.run(self)

    def build_extension(self, ext):
        try:
            extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

            # required for auto-detection & inclusion of auxiliary "native" libs

            debug = (
                int(os.environ.get("DEBUG", 0)) if self.debug is None else self.debug
            )
            cfg = "Debug" if debug else "Release"

            # CMake lets you override the generator - we need to check this.
            # Can be set with Conda-Build, for example.
            cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

            # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
            # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
            # from Python.
            cmake_args = [
                f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}",
                f"-DPYTHON_EXECUTABLE={sys.executable}",
                f"-DCMAKE_BUILD_TYPE={cfg}",  # not used on MSVC, but no harm
            ]
            build_args = []
            # Adding CMake arguments set as environment variable
            # (needed e.g. to build for ARM OSx on conda-forge)
            if "CMAKE_ARGS" in os.environ:
                cmake_args += [
                    item for item in os.environ["CMAKE_ARGS"].split(" ") if item
                ]

            # In this example, we pass in the version to C++. You might not need to.
            cmake_args += [f"-DEXAMPLE_VERSION_INFO={self.distribution.get_version()}"]

            if self.compiler.compiler_type != "msvc":
                # Using Ninja-build since it a) is available as a wheel and b)
                # multithreads automatically. MSVC would require all variables be
                # exported for Ninja to pick it up, which is a little tricky to do.
                # Users can override the generator with CMAKE_GENERATOR in CMake
                # 3.15+.
                if not cmake_generator:
                    try:
                        import ninja  # noqa: F401

                        cmake_args += ["-GNinja"]
                    except ImportError:
                        pass

            else:

                # Single config generators are handled "normally"
                single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})

                # CMake allows an arch-in-generator style for backward compatibility
                contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})

                # Specify the arch if using MSVC generator, but only if it doesn't
                # contain a backward-compatibility arch spec already in the
                # generator name.
                if not single_config and not contains_arch:
                    cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

                # Multi-config generators have a different way to specify configs
                if not single_config:
                    cmake_args += [
                        f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{cfg.upper()}={extdir}"
                    ]
                    build_args += ["--config", cfg]

            if sys.platform.startswith("darwin"):
                # Cross-compile support for macOS - respect ARCHFLAGS if set
                archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
                if archs:
                    cmake_args += [
                        "-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))
                    ]

            # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
            # across all generators.
            if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
                # self.parallel is a Python 3 only way to set parallel jobs by hand
                # using -j in the build_ext call, not supported by pip or PyPA-build.
                if hasattr(self, "parallel") and self.parallel:
                    # CMake 3.12+ only.
                    build_args += [f"-j{self.parallel}"]

            if not os.path.exists(self.build_temp):
                os.makedirs(self.build_temp)

            print(["cmake", ext.sourcedir] + cmake_args)
            subprocess.check_call(
                ["cmake", ext.sourcedir] + cmake_args, cwd=self.build_temp
            )
            print(["cmake", "--build", "."] + build_args)
            subprocess.check_call(
                ["cmake", "--build", "."] + build_args, cwd=self.build_temp
            )
        except Exception as e:
            traceback.print_stack()
            print(e)
            print(f"c_tools module fail to compile")


with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    "pandas",
    "numpy",
    "aiosmtplib",
    "python-dateutil",
    "pymongo",
    "redis",
    "pyyaml",
    "motor",
]

extras_require = {
    "full": ["luigi", "scikit-learn", "joblib"],
}


setup(
    name="lightning_fast",
    version="0.0.96",
    author="yingxiongxqs",
    author_email="yingxiongxqs@126.com",
    description="Fast Tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/xuqiushi/lightning_fast",
    # cmdclass={
    #     "install": CustomInstallCommand,
    #     # 'develop': CustomDevelopCommand,
    #     "egg_info": CustomEggInfoCommand,
    # },
    ext_package="lightning_fast.c_tools",
    ext_modules=[CMakeExtension("encoders", "c_source")],
    cmdclass={"build_ext": CMakeBuild},
    # package_dir={"": "lightning_fast"},
    packages=find_packages(exclude=("tests",)),
    install_requires=install_requires,
    include_package_data=True,
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Text Processing",
    ],
    python_requires=">=3.7",
    zip_safe=False,
    extras_require=extras_require,
)



import logging
import pathlib
from logging import handlers
from typing import Optional


class LoggerUtil(object):
    """
    使用此类生成一个简单的logger。此logger可以将log分文件打印并保存。
    """

    @classmethod
    def get_simple_logger(
        cls,
        log_path: Optional[pathlib.Path] = None,
        log_file_name: Optional[str] = None,
        logger_level: Optional[int] = None,
    ) -> logging.Logger:
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        logger_level = logging.INFO if not logger_level else logger_level
        if not log_path or not log_file_name:
            logger = logging.getLogger()
        else:
            # log_path.mkdir(parents=True, exist_ok=True)
            logger = logging.getLogger(log_file_name)
            fh = handlers.RotatingFileHandler(
                log_path / f"{log_file_name}.log", maxBytes=10240 * 1000, backupCount=5,
            )
            fh.setLevel(logging.INFO)
            fh.setFormatter(formatter)
            logger.addHandler(fh)
        logger.setLevel(logger_level)
        return logger



import luigi
from luigi.contrib.hive import HivePartitionTarget


class HiveDtPartitionCheck(luigi.Task):
    """
    事实上在另一个地方: offline/variant_data_indicators/base_components/util_tasks.py
    中有类似的一个通用检查hive partition的任务。但是那里的这个类耦合比较严重, 重构留待之后
    同时继承此类与TaskBase的实现类即可

    """

    hive_path = luigi.Parameter()
    partition = luigi.Parameter()

    def run(self):
        existed = HivePartitionTarget(
            str(self.hive_path), partition={"dt": self.partition}
        ).exists()
        if existed:
            with self.output().open("w") as f:
                f.write(f"Got f{self.partition} of {self.hive_path}。")
        else:
            raise ValueError(f"Got no f{self.partition} of {self.hive_path}。")